"""Main class for the Crownstone cloud lib."""
import hashlib
import logging
import asyncio
import aiohttp
from typing import Optional
from crownstone_cloud._RequestHandlerInstance import RequestHandler
from crownstone_cloud.lib.cloudModels.spheres import Spheres
from crownstone_cloud.lib.cloudModels.crownstones import Crownstone

_LOGGER = logging.getLogger(__name__)


class CrownstoneCloud:
    """Create a Crownstone cloud instance."""

    def __init__(
            self,
            email: str,
            password: str,
            loop: asyncio.AbstractEventLoop = None,
            websession: aiohttp.ClientSession = None
    ) -> None:
        self.login_data = {'email': email, 'password': self.password_to_hash(password)}
        self.loop = loop or asyncio.get_event_loop()
        # set request handler params
        RequestHandler.websession = websession or aiohttp.ClientSession(loop=loop)
        RequestHandler.login_data = self.login_data
        # data
        self.spheres: Optional[Spheres] = None

    async def async_initialize(self) -> None:
        """
        Initialize the cloud data.

        This method is a coroutine.
        """
        if RequestHandler.access_token is None:
            await self.async_login()
        await self.async_synchronize()

    def initialize(self) -> None:
        """Initialize the cloud data."""
        self.loop.run_until_complete(self.async_initialize())

    @staticmethod
    def set_access_token(access_token: str) -> None:
        RequestHandler.access_token = access_token

    @staticmethod
    def get_access_token() -> str:
        return RequestHandler.access_token

    async def async_login(self) -> None:
        """
        Login to Crownstone API.

        This method is a coroutine.
        """
        result = await RequestHandler.post('users', 'login', json=self.login_data)

        # Set access token & user id
        RequestHandler.access_token = result['id']
        self.spheres = Spheres(self.loop, result['userId'])

        _LOGGER.info("Login to Crownstone Cloud successful")

    async def async_synchronize(self) -> None:
        """
        Sync all data from cloud.

        This method is a coroutine.
        """
        _LOGGER.info("Initiating all cloud data")
        # get the sphere data
        await self.spheres.async_update_sphere_data()

        # get the data from the sphere attributes
        for sphere in self.spheres:
            await asyncio.gather(
                sphere.async_update_sphere_presence(),
                sphere.crownstones.async_update_crownstone_data(),
                sphere.locations.async_update_location_data(),
                sphere.locations.async_update_location_presence(),
                sphere.users.async_update_user_data()
            )
        _LOGGER.info("Cloud data successfully initialized")

    def get_crownstone(self, crownstone_name) -> Crownstone:
        """Get a crownstone by name without specifying a sphere."""
        for sphere in self.spheres:
            for crownstone in sphere.crownstones:
                if crownstone.name == crownstone_name:
                    return crownstone

    def get_crownstone_by_id(self, crownstone_id) -> Crownstone:
        """Get a crownstone by id without specifying a sphere."""
        for sphere in self.spheres:
            return sphere.crownstones[crownstone_id]

    @staticmethod
    def reset() -> None:
        """Cleanup the request handler instance data."""
        RequestHandler.access_token = None
        RequestHandler.login_data = None

    @staticmethod
    async def async_close_session() -> None:
        """
        Close the websession after we are done.

        This method is a coroutine.
        """
        await RequestHandler.websession.close()
        _LOGGER.info("Session closed.")

    def close_session(self) -> None:
        """Close the websession after we are done."""
        self.loop.run_until_complete(self.async_close_session())

    @staticmethod
    def password_to_hash(password):
        """Generate a sha1 password from string."""
        if password is None:
            return None
        pw_hash = hashlib.sha1(password.encode('utf-8'))
        return pw_hash.hexdigest()



"""Sphere handler for Crownstone cloud data."""
from crownstone_cloud._RequestHandlerInstance import RequestHandler
from crownstone_cloud.lib.cloudModels.crownstones import Crownstones
from crownstone_cloud.lib.cloudModels.locations import Locations
from crownstone_cloud.lib.cloudModels.users import Users
import asyncio


class Spheres:
    """Handler for the spheres of the user."""

    def __init__(self, loop: asyncio.AbstractEventLoop, user_id: str) -> None:
        """Initialization."""
        self.loop = loop
        self.spheres = {}
        self.user_id = user_id

    def __iter__(self):
        """Iterate over spheres."""
        return iter(self.spheres.values())

    async def async_update_sphere_data(self) -> None:
        """
        Get the spheres for the user from the cloud.

        This method is a coroutine.
        """
        sphere_data = await RequestHandler.get(
            'users', 'spheres', model_id=self.user_id
        )
        # process items
        removed_items = []
        new_items = []
        for sphere in sphere_data:
            sphere_id = sphere['id']
            exists = self.spheres.get(sphere_id)
            # check if the Sphere already exists
            # it is important that we don't throw away existing objects, as they need to remain functional
            if exists:
                # update data
                self.spheres[sphere_id].data = sphere
            else:
                # add new Sphere
                self.spheres[sphere_id] = Sphere(self.loop, sphere, self.user_id)

            # generate list with new id's to check with the existing id's
            new_items.append(sphere_id)

        # check for removed items
        for sphere_id in self.spheres:
            if sphere_id not in new_items:
                removed_items.append(sphere_id)

        # remove items from dict
        for sphere_id in removed_items:
            del self.spheres[sphere_id]

    def update_sphere_data(self) -> None:
        """Get the sphere data from the cloud."""
        self.loop.run_until_complete(self.async_update_sphere_data())

    def find(self, sphere_name: str) -> object or None:
        """Search for a sphere by name and return sphere object if found."""
        for sphere in self.spheres.values():
            if sphere_name == sphere.name:
                return sphere

        return None

    def find_by_id(self, sphere_id: str) -> object or None:
        """Search for a sphere by id and return sphere object if found."""
        return self.spheres[sphere_id]


class Sphere:
    """Represents a Sphere"""

    def __init__(self, loop: asyncio.AbstractEventLoop, data: dict, user_id: str):
        """Initialization."""
        self.loop = loop
        self.data = data
        self.user_id = user_id
        self.crownstones = Crownstones(loop, self.cloud_id)
        self.locations = Locations(loop, self.cloud_id)
        self.users = Users(loop, self.cloud_id)
        self.keys: dict = {}
        self.present_people = []

    @property
    def name(self) -> str:
        """Returns the name of this Sphere."""
        return self.data['name']

    @property
    def cloud_id(self) -> str:
        """Return the cloud id of this Sphere."""
        return self.data['id']

    @property
    def unique_id(self) -> int:
        """Return the unique id of this Sphere."""
        return self.data['uid']

    async def async_get_keys(self) -> dict:
        """
        Get the user keys for this sphere, that can be used for BLE (optional).

        This method is a coroutine.
        """
        keys = await RequestHandler.get('users', 'keysV2', model_id=self.user_id)
        for key_set in keys:
            if key_set['sphereId'] == self.cloud_id:
                for keyType in key_set['sphereKeys']:
                    self.keys[keyType['keyType']] = keyType['key']

        return self.keys

    async def async_update_sphere_presence(self) -> None:
        """
        Replaces the current presence with that of the cloud.

        This method is a coroutine.
        """
        self.present_people = []
        presence_data = await RequestHandler.get('Spheres', 'presentPeople', model_id=self.cloud_id)
        for user in presence_data:
            self.present_people.append(user['userId'])

    def get_keys(self) -> dict:
        """Get the user keys for this Sphere, that can be used for BLE (optional)."""
        return self.loop.run_until_complete(self.async_get_keys())



"""Exceptions for crownstone lib connection lib"""


AuthError = {
    'LOGIN_FAILED': 'Wrong email or password provided',
    'USERNAME_EMAIL_REQUIRED': 'Email or password not provided',
    'LOGIN_FAILED_EMAIL_NOT_VERIFIED': 'Email has not been verified, please do that first',
}


class CrownstoneAuthenticationError(Exception):
    """Raised when authentication with API ended in error"""
    type = None
    message = None

    def __init__(self, type, message=None):
        self.type = type
        self.message = message


class CrownstoneUnknownError(Exception):
    """Raised when the error is not known / no data"""



"""Handler for requests to the Crownstone cloud"""
import logging
import json
from urllib.parse import quote
from aiohttp import ClientSession
from typing import Any, Optional
from crownstone_cloud.const import BASE_URL
from crownstone_cloud.exceptions import (
    CrownstoneAuthenticationError,
    CrownstoneUnknownError,
    AuthError
)

_LOGGER = logging.getLogger(__name__)


class RequestHandler:
    """Handles requests to the Crownstone lib."""

    def __init__(self) -> None:
        self.access_token: Optional[str] = None
        self.websession: Optional[ClientSession] = None
        self.login_data: Optional[dict] = None

    async def post(
            self,
            model: str,
            endpoint: str,
            model_id: str = None,
            json: dict = None
    ) -> dict:
        """
        Post request

        :param model: model type. users, spheres, stones, locations, devices.
        :param endpoint: endpoints. e.g. spheres, keys, presentPeople.
        :param model_id: required id for the endpoint. e.g. userId for users, sphereId for spheres.
        :param json: Dictionary with the data that should be posted.
        :return: Dictionary with the response from the lib.
        """
        if self.access_token is None:
            url = f'{BASE_URL}{model}/{endpoint}'
        elif model_id:
            url = f'{BASE_URL}{model}/{model_id}/{endpoint}?access_token={self.access_token}'
        else:
            url = f'{BASE_URL}{model}{endpoint}?access_token={self.access_token}'

        return await self.request('post', url, json)

    async def get(
            self,
            model: str,
            endpoint: str,
            filter: dict = None,
            model_id: str = None
    ) -> dict:
        """
        Get request

        :param model: model type. users, spheres, stones, locations, devices.
        :param endpoint: endpoints. e.g. spheres, keys, presentPeople.
        :param filter: filter output or add extra data to output.
        :param model_id: required id for the endpoint. e.g. userId for users, sphereId for spheres.
        :return: Dictionary with the response from the lib.
        """
        if filter and model_id:
            url = f'{BASE_URL}{model}/{model_id}/{endpoint}?filter={self.quote_json(filter)}&access_token={self.access_token}'
        elif model_id and not filter:
            url = f'{BASE_URL}{model}/{model_id}/{endpoint}?access_token={self.access_token}'
        else:
            url = f'{BASE_URL}{model}{endpoint}?access_token={self.access_token}'

        return await self.request('get', url)

    async def put(
            self,
            model: str,
            endpoint: str,
            model_id: str,
            command: str,
            value: Any
    ) -> dict:
        """
        Put request

        :param model: model type. users, spheres, stones, locations, devices.
        :param endpoint: endpoints. e.g. spheres, keys, presentPeople.
        :param model_id: required id for the endpoint. e.g. userId for users, sphereId for spheres.
        :param command: used for command requests. e.g. 'switchState'.
        :param value: the value to be put for the command. e.g 'switchState', 1
        :return: Dictionary with the response from the lib.
        """
        url = f'{BASE_URL}{model}/{model_id}/{endpoint}?{command}={str(value)}&access_token={self.access_token}'

        return await self.request('put', url)

    async def request(self, method: str, url: str, json: dict = None) -> dict:
        """Make request and check data for errors"""
        async with self.websession.request(method, url, json=json) as result:
            data = await result.json()
            refresh = await self.raise_on_error(data)
            if refresh:
                new_url = url.replace(url.split('access_token=', 1)[1], self.access_token)
                await self.request(method, new_url, json=json)
            return data

    async def raise_on_error(self, data) -> bool:
        """Check for error message"""
        if isinstance(data, dict) and 'error' in data:
            error = data['error']

            if 'code' in error:
                error_type = error['code']
                try:
                    if error_type == 'INVALID_TOKEN' or error_type == 'AUTHORIZATION_REQUIRED':
                        _LOGGER.warning("Token expired. Refreshing now...")
                        await self.refresh_token()
                        return True  # re-run the request
                    else:
                        for type, message in AuthError.items():
                            if type == error_type:
                                raise CrownstoneAuthenticationError(type, message)
                except ValueError:
                    raise CrownstoneUnknownError("Unknown error occurred.")
            else:
                _LOGGER.error(error['message'])

        return False

    async def refresh_token(self):
        self.access_token = None
        response = await self.post('users', 'login', json=self.login_data)
        self.access_token = response['id']

    @staticmethod
    def quote_json(_json: dict) -> str:
        stringified = json.dumps(_json)
        return quote(stringified)



"""Crownstone handler for Crownstone cloud data"""
from crownstone_cloud._RequestHandlerInstance import RequestHandler
from crownstone_cloud.const import (
    DIMMING_ABILITY
)
import logging
import asyncio

_LOGGER = logging.Logger(__name__)


class Crownstones:
    """Handler for the crownstones of a sphere."""

    def __init__(self, loop: asyncio.AbstractEventLoop, sphere_id: str) -> None:
        """Initialization."""
        self.loop = loop
        self.crownstones = {}
        self.sphere_id = sphere_id

    def __iter__(self):
        """Iterate over crownstones."""
        return iter(self.crownstones.values())

    async def async_update_crownstone_data(self) -> None:
        """Get the crownstones data from the cloud."""
        # include abilities and current switch state in the request
        data_filter = {"include": ["currentSwitchState", {"abilities": "properties"}]}
        # request data
        crownstone_data = await RequestHandler.get(
            'Spheres', 'ownedStones', filter=data_filter, model_id=self.sphere_id
        )
        # process items
        removed_items = []
        new_items = []
        for crownstone in crownstone_data:
            crownstone_id = crownstone['id']
            exists = self.crownstones.get(crownstone_id)
            # check if the crownstone already exists
            # it is important that we don't throw away existing objects, as they need to remain functional
            if exists:
                # update data and update abilities
                self.crownstones[crownstone_id].data = crownstone
            else:
                # add new Crownstone
                self.crownstones[crownstone_id] = Crownstone(self.loop, crownstone)

            # update the abilities of the Crownstone from the data
            self.crownstones[crownstone_id].update_abilities()

            # generate list with new id's to check with the existing id's
            new_items.append(crownstone_id)

        # check for removed items
        for crownstone_id in self.crownstones:
            if crownstone_id not in new_items:
                removed_items.append(crownstone_id)

        # remove items from dict
        for crownstone_id in removed_items:
            del self.crownstones[crownstone_id]

    def update_crownstone_data(self) -> None:
        """Sync function for updating the crownstone data."""
        self.loop.run_until_complete(self.async_update_crownstone_data())

    def find(self, crownstone_name: str) -> object or None:
        """Search for a crownstone by name and return crownstone object if found."""
        for crownstone in self.crownstones.values():
            if crownstone_name == crownstone.name:
                return crownstone

    def find_by_id(self, crownstone_id) -> object or None:
        """Search for a crownstone by id and return crownstone object if found."""
        return self.crownstones[crownstone_id]

    def find_by_uid(self, crownstone_uid) -> object or None:
        """Search for a crownstone by uid and return crownstone object if found."""
        for crownstone in self.crownstones.values():
            if crownstone_uid == crownstone.unique_id:
                return crownstone


class CrownstoneAbility:
    """Represents a Crownstone Ability"""

    def __init__(self, data: dict) -> None:
        """Initialization"""
        self.data = data
        self.is_enabled = self.data['enabled']
        self.properties = self.data['properties']

    @property
    def type(self) -> str:
        """Return the ability type."""
        return self.data['type']

    @property
    def ability_id(self) -> str:
        """Return the ability id."""
        return self.data['id']

    @property
    def crownstone_id(self) -> str:
        """Return the Crownstone id."""
        return self.data['stoneId']


class Crownstone:
    """Represents a Crownstone"""

    def __init__(self, loop: asyncio.AbstractEventLoop, data: dict) -> None:
        """Initialization."""
        self.loop = loop
        self.data = data
        self.abilities = {}

    @property
    def name(self) -> str:
        """Return the name of this Crownstone."""
        return self.data['name']

    @property
    def unique_id(self) -> int:
        """Return the unique_id of this Crownstone."""
        return self.data['uid']

    @property
    def cloud_id(self) -> str:
        """Return the cloud id of this Crownstone."""
        return self.data['id']

    @property
    def type(self) -> str:
        """Return the Crownstone type."""
        return self.data['type']

    @property
    def sw_version(self) -> str:
        """Return the firmware version of this Crownstone."""
        return self.data['firmwareVersion']

    @property
    def icon(self) -> str:
        """Return the icon of this Crownstone."""
        return self.data['icon']

    @property
    def state(self) -> float:
        """Return the last reported state for this Crownstone."""
        return self.data['currentSwitchState']['switchState']

    @state.setter
    def state(self, value: float) -> None:
        """Set a new state for this Crownstone."""
        self.data['currentSwitchState']['switchState'] = value

    def update_abilities(self) -> None:
        """Add/update the abilities for this Crownstone."""
        for ability in self.data['abilities']:
            self.abilities[ability['type']] = CrownstoneAbility(ability)

    async def async_turn_on(self) -> None:
        """
        Turn this crownstone on.

        This method is a coroutine.
        """
        await RequestHandler.put(
            'Stones', 'setSwitchStateRemotely', model_id=self.cloud_id, command='switchState', value=1
        )

    async def async_turn_off(self) -> None:
        """
        Turn this crownstone off.

        This method is a coroutine.
        """
        await RequestHandler.put(
            'Stones', 'setSwitchStateRemotely', model_id=self.cloud_id, command='switchState', value=0
        )

    async def async_set_brightness(self, brightness: float) -> None:
        """
        Set the brightness of this crownstone, if dimming enabled.

        :param brightness: brightness value between (0 - 1)

        This method is a coroutine.
        """
        if self.abilities[DIMMING_ABILITY].is_enabled:
            if brightness < 0 or brightness > 1:
                raise ValueError("Enter a value between 0 and 1")
            else:
                await RequestHandler.put(
                    'Stones', 'setSwitchStateRemotely', model_id=self.cloud_id, command='switchState', value=brightness
                )
        else:
            _LOGGER.warning("Dimming is not enabled for this crownstone. Go to the crownstone app to enable it")

    def turn_on(self) -> None:
        """Turn this Crownstone on."""
        self.loop.run_until_complete(self.async_turn_on())

    def turn_off(self) -> None:
        """Turn this Crownstone off."""
        self.loop.run_until_complete(self.async_turn_off())

    def set_brightness(self, brightness: float) -> None:
        """
        Set the brightness of this crownstone, if dimming enabled.

        :param brightness: the brightness value between (0 - 1)
        """
        self.loop.run_until_complete(self.async_set_brightness(brightness))
